import pytest
import functools


def skip_if_import_exception(function):
    """Assist in skipping tests failing because of missing dependencies."""
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except ImportError as err:
            pytest.skip(str(err))
    return wrapper


def start_http_server(redirect=None):
    """Returns the port of the newly started HTTP server."""
    def _do_redirect(handler, location):
        handler.send_response(301)
        handler.send_header("Location", location)
        handler.end_headers()

    import socket
    import sys
    import threading
    # Start HTTP server to serve TAR files.
    # pylint:disable=g-import-not-at-top
    if sys.version_info[0] == 2:
        import BaseHTTPServer
        import SimpleHTTPServer

        class HTTPServerV6(BaseHTTPServer.HTTPServer):

            address_family = socket.AF_INET6

        class RedirectHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):

            def do_GET(self):
                _do_redirect(self, redirect)

        server = HTTPServerV6(("", 0), RedirectHandler if redirect else
                              SimpleHTTPServer.SimpleHTTPRequestHandler)
        server_port = server.server_port
    else:
        import http.server
        import socketserver

        class TCPServerV6(socketserver.TCPServer):

            address_family = socket.AF_INET6

        class RedirectHandler(http.server.SimpleHTTPRequestHandler):

            def do_GET(self):
                _do_redirect(self, redirect)

        server = TCPServerV6(("", 0), RedirectHandler if redirect else
                             http.server.SimpleHTTPRequestHandler)
        _, server_port, _, _ = server.server_address
    # pylint:disable=g-import-not-at-top

    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()

    return server_port



import os
import shutil
import unittest
import numpy as np
from .helper import skip_if_import_exception, start_http_server

from signaturizer.exporter import export_smilespred, export_savedmodel
from signaturizer import Signaturizer


class TestSignaturizer(unittest.TestCase):

    def setUp(self):
        # path for test data
        test_dir = os.path.dirname(os.path.realpath(__file__))
        self.data_dir = os.path.join(test_dir, 'data')
        self.tmp_dir = os.path.join(test_dir, 'tmp')
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)
        os.mkdir(self.tmp_dir)
        os.mkdir(os.path.join(self.tmp_dir, 'vXXX'))
        self.cwd = os.getcwd()
        os.chdir(self.tmp_dir)
        self.server_port = start_http_server()
        self.test_smiles = [
            # Erlotinib
            'COCCOC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC=CC(=C3)C#C)OCCOC',
            # Diphenhydramine
            'CN(C)CCOC(C1=CC=CC=C1)C2=CC=CC=C2'
        ]

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)
            pass

    @skip_if_import_exception
    def test_export(self):
        # export smilespred
        version = 'vXXX'
        module_file = 'dest_smilespred.tar.gz'
        module_destination = os.path.join(
            self.tmp_dir, version, module_file)
        tmp_path_smilespred = os.path.join(self.tmp_dir, 'export_smilespred')
        smilespred_path = os.path.join(self.data_dir, 'models', 'smiles')
        export_smilespred(smilespred_path, module_destination,
                          tmp_path=tmp_path_smilespred, clear_tmp=False)
        base_url = "http://localhost:%d/" % (self.server_port)
        module = Signaturizer(module_file, base_url=base_url, version=version)
        res = module.predict(self.test_smiles)
        pred = res.signature[:]
        ref_pred_file = os.path.join(
            self.data_dir, 'models', 'smiles_pred.npy')
        #np.save(ref_pred_file, pred)
        pred_ref = np.load(ref_pred_file)
        np.testing.assert_almost_equal(pred_ref, pred)

    @skip_if_import_exception
    def test_export_applicability(self):
        # export smilespred and applicability
        version = 'vXXX'
        module_file = 'dest_smilespred.tar.gz'
        module_destination = os.path.join(
            self.tmp_dir, version, module_file)
        tmp_path_smilespred = os.path.join(self.tmp_dir, 'export_smilespred')
        smilespred_path = os.path.join(self.data_dir, 'models', 'smiles')
        apppred_path = os.path.join(self.data_dir, 'models', 'applicability')
        export_smilespred(smilespred_path, module_destination,
                          tmp_path=tmp_path_smilespred, clear_tmp=False,
                          applicability_path=apppred_path)
        base_url = "http://localhost:%d/" % (self.server_port)
        module = Signaturizer(module_file, base_url=base_url, version=version,
                              applicability=True)
        res = module.predict(self.test_smiles)
        pred = res.signature[:]
        ref_pred_file = os.path.join(
            self.data_dir, 'models', 'smiles_pred.npy')
        pred_ref = np.load(ref_pred_file)
        np.testing.assert_almost_equal(pred_ref, pred)
        apred = res.applicability[:]
        ref_apred_file = os.path.join(
            self.data_dir, 'models', 'applicability_pred.npy')
        #np.save(ref_apred_file, apred)
        apred_ref = np.load(ref_apred_file)
        np.testing.assert_almost_equal(apred_ref, apred)

    @skip_if_import_exception
    def test_export_consistency(self):
        """Compare the exported module to the original SMILES predictor.

        N.B. This test is working only with a valid CC instance available.
        """
        from chemicalchecker import ChemicalChecker
        from chemicalchecker.core.signature_data import DataSignature

        # load CC instance and smiles prediction model
        cc = ChemicalChecker()
        s3 = cc.signature('B1.001', 'sign3')
        tmp_pred_ref = os.path.join(self.tmp_dir, 'tmp.h5')
        s3.predict_from_smiles(self.test_smiles, tmp_pred_ref)
        pred_ref = DataSignature(tmp_pred_ref)[:]

        # export smilespred
        version = 'vXXX'
        module_file = 'dest_smilespred.tar.gz'
        module_destination = os.path.join(
            self.tmp_dir, version, module_file)
        tmp_path_smilespred = os.path.join(self.tmp_dir, 'export_smilespred')
        export_smilespred(
            os.path.join(s3.model_path, 'smiles_final'),
            module_destination, tmp_path=tmp_path_smilespred, clear_tmp=False)
        # test intermediate step
        module = Signaturizer(tmp_path_smilespred, local=True)
        res = module.predict(self.test_smiles)
        pred = res.signature[:]
        np.testing.assert_almost_equal(pred_ref, pred)
        # test final step
        base_url = "http://localhost:%d/" % (self.server_port)
        module = Signaturizer(module_file, base_url=base_url, version=version)
        res = module.predict(self.test_smiles)
        pred = res.signature[:]
        np.testing.assert_almost_equal(pred_ref, pred)

        # export savedmodel
        module_destination = os.path.join(
            self.tmp_dir, 'dest_savedmodel.tar.gz')
        tmp_path_savedmodel = os.path.join(self.tmp_dir, 'export_savedmodel')
        export_savedmodel(
            tmp_path_smilespred, module_destination,
            tmp_path=tmp_path_savedmodel, clear_tmp=False)
        # test intermediate step
        module = Signaturizer(tmp_path_savedmodel, local=True)
        res = module.predict(self.test_smiles)
        pred = res.signature[:]
        np.testing.assert_almost_equal(pred_ref, pred)
        # test final step
        module = Signaturizer(module_file, base_url=base_url, version=version)
        res = module.predict(self.test_smiles)
        pred = res.signature[:]
        np.testing.assert_almost_equal(pred_ref, pred)



import os
import math
import pickle
import shutil
import unittest
import numpy as np

from signaturizer import Signaturizer


class TestSignaturizer(unittest.TestCase):

    def setUp(self):
        # path for test data
        test_dir = os.path.dirname(os.path.realpath(__file__))
        self.data_dir = os.path.join(test_dir, 'data')
        self.tmp_dir = os.path.join(test_dir, 'tmp')
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)
        os.mkdir(self.tmp_dir)
        self.test_smiles = [
            # Erlotinib
            'COCCOC1=C(C=C2C(=C1)C(=NC=N2)NC3=CC=CC(=C3)C#C)OCCOC',
            # Diphenhydramine
            'CN(C)CCOC(C1=CC=CC=C1)C2=CC=CC=C2'
        ]
        self.invalid_smiles = ['C', 'C&', 'C']
        self.tautomer_smiles = ['CC(O)=Nc1ccc(O)cc1', 'CC(=O)NC1=CC=C(C=C1)O']
        self.inchi = [
            'InChI=1S/C8H9NO2/c1-6(10)9-7-2-4-8(11)5-3-7/h2-5,11H,1H3,(H,9,10)'
        ]

    def tearDown(self):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir, ignore_errors=True)
            pass

    def test_predict(self):
        # load reference predictions
        ref_file = os.path.join(self.data_dir, 'pred.pkl')
        pred_ref = pickle.load(open(ref_file, 'rb'))
        # load module and predict
        module_dir = os.path.join(self.data_dir, 'B1')
        module = Signaturizer(module_dir, local=True)
        res = module.predict(self.test_smiles)
        np.testing.assert_almost_equal(pred_ref, res.signature[:])
        # test saving to file
        destination = os.path.join(self.tmp_dir, 'pred.h5')
        res = module.predict(self.test_smiles, destination)
        self.assertTrue(os.path.isfile(destination))
        np.testing.assert_almost_equal(pred_ref, res.signature[:])
        # test prediction of invalid SMILES
        res = module.predict(self.invalid_smiles)
        for comp in res.signature[0]:
            self.assertFalse(math.isnan(comp))
        for comp in res.signature[1]:
            self.assertTrue(math.isnan(comp))
        for comp in res.signature[2]:
            self.assertFalse(math.isnan(comp))

    def test_predict_multi(self):
        # load multiple model and check that results stacked correctly
        module_dirs = list()
        A1_path = os.path.join(self.data_dir, 'A1')
        B1_path = os.path.join(self.data_dir, 'B1')
        module_dirs.append(A1_path)
        module_dirs.append(B1_path)
        module_A1B1 = Signaturizer(module_dirs, local=True)
        res_A1B1 = module_A1B1.predict(self.test_smiles)
        self.assertEqual(res_A1B1.signature.shape[0], 2)
        self.assertEqual(res_A1B1.signature.shape[1], 128 * 2)

        module_A1 = Signaturizer(A1_path, local=True)
        res_A1 = module_A1.predict(self.test_smiles)
        np.testing.assert_almost_equal(res_A1B1.signature[:, :128],
                                       res_A1.signature)

        module_B1 = Signaturizer(B1_path, local=True)
        res_B1 = module_B1.predict(self.test_smiles)
        np.testing.assert_almost_equal(res_A1B1.signature[:, 128:],
                                       res_B1.signature)

        res = module_A1B1.predict(self.invalid_smiles)
        for comp in res.signature[0]:
            self.assertFalse(math.isnan(comp))
        for comp in res.signature[1]:
            self.assertTrue(math.isnan(comp))
        for comp in res.signature[2]:
            self.assertFalse(math.isnan(comp))
        self.assertTrue(all(res.failed == [False, True, False]))

    def test_predict_global_remote(self):
        module = Signaturizer(['GLOBAL'])
        res = module.predict(self.test_smiles)
        self.assertEqual(res.signature.shape[0], 2)
        self.assertEqual(res.signature.shape[1], 128 * 25)

    def test_overwrite(self):
        module_dir = os.path.join(self.data_dir, 'B1')
        module = Signaturizer(module_dir, local=True)
        destination = os.path.join(self.tmp_dir, 'pred.h5')
        module.predict(self.test_smiles, destination)
        # repeating writing will result in an exception
        with self.assertRaises(Exception):
            module.predict(self.test_smiles, destination)

    def test_tautomers(self):
        module = Signaturizer('A1')
        res = module.predict(self.tautomer_smiles)
        self.assertTrue(all(res.signature[0] == res.signature[1]))

    def test_inchi(self):
        module = Signaturizer('A1')
        res_inchi = module.predict(self.inchi, keytype='InChI')
        res_smiles = module.predict([self.tautomer_smiles[0]])
        self.assertTrue(all(res_inchi.signature[0] == res_smiles.signature[0]))

    def test_all_single(self):
        module = Signaturizer('A1')
        res_all = module.predict(self.test_smiles)
        for idx, smile in enumerate(self.test_smiles):
            res_single = module.predict([smile])
            self.assertTrue(
                all(res_all.signature[idx] == res_single.signature[0]))



import os
import shutil
import tempfile
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    import tensorflow.compat.v1 as tf
    import tensorflow_hub as hub


def export_smilespred(smilespred_path, destination,
                      tmp_path=None, clear_tmp=True, compress=True,
                      applicability_path=None):
    """Export our Keras Smiles predictor to the TF-hub module format."""
    from keras import backend as K
    from chemicalchecker.tool.smilespred import Smilespred
    from chemicalchecker.tool.smilespred import ApplicabilityPredictor

    if tmp_path is None:
        tmp_path = tempfile.mkdtemp()
    # save to savedmodel format
    with tf.Graph().as_default():
        smilespred = Smilespred(smilespred_path)
        smilespred.build_model(load=True)
        model = smilespred.model
        '''
        with tf.Session() as sess:
            sess.run(tf.tables_initializer())
            sess.run(tf.global_variables_initializer())
            tf.saved_model.simple_save(
                sess,
                tmp_path,
                inputs={'default': model.input},
                outputs={'default': model.output}
            )
        '''
        signature = tf.saved_model.signature_def_utils.predict_signature_def(
            inputs={'default': model.input}, outputs={'default': model.output})
        signature_def_map = {'serving_default': signature}
        if applicability_path is not None:
            app_pred = ApplicabilityPredictor(applicability_path)
            app_pred.build_model(load=True)
            app_model = app_pred.model
            signature_app = tf.saved_model.signature_def_utils.predict_signature_def(
                inputs={'default': app_model.input},
                outputs={'default': app_model.output})
            signature_def_map.update({'applicability': signature_app})
        if tmp_path is None:
            tmp_path = tempfile.mkdtemp()
        builder = tf.saved_model.builder.SavedModelBuilder(tmp_path)
        builder.add_meta_graph_and_variables(
            sess=K.get_session(),
            tags=['serve'],
            signature_def_map=signature_def_map)
        builder.save()
    # now export savedmodel to module
    export_savedmodel(tmp_path, destination, compress=compress)
    # clean temporary folder
    if clear_tmp:
        shutil.rmtree(tmp_path)


def export_savedmodel(savedmodel_path, destination,
                      tmp_path=None, clear_tmp=True, compress=True):
    """Export Tensorflow SavedModel to the TF-hub module format."""
    if tmp_path is None:
        tmp_path = tempfile.mkdtemp()
    # save to hub module format
    with tf.Graph().as_default():
        spec = hub.create_module_spec_from_saved_model(savedmodel_path)
        module = hub.Module(spec, tags=['serve'])
        with tf.Session() as sess:
            sess.run(tf.tables_initializer())
            sess.run(tf.global_variables_initializer())
            module.export(tmp_path, sess)
    if compress:
        # compress the exported files to destination
        os.system("tar -cz -f %s --owner=0 --group=0 -C %s ." %
                  (destination, tmp_path))
    else:
        shutil.copytree(tmp_path, destination)
    # clean temporary folder
    if clear_tmp:
        shutil.rmtree(tmp_path)


def export_batch(cc, destination_dir, datasets=None, applicability=True):
    """Export all CC Smiles predictor to the TF-hub module format."""
    if datasets is None:
        datasets = cc.datasets_exemplary()
    for ds in datasets:
        s3 = cc.signature(ds, 'sign3')
        pred_path = os.path.join(s3.model_path, 'smiles_final')
        mdl_dest = os.path.join(destination_dir, ds[:2])
        if applicability:
            apred_path = os.path.join(
                s3.model_path, 'smiles_applicability_final')
            export_smilespred(pred_path, mdl_dest, compress=False,
                              applicability_path=apred_path)
            export_smilespred(pred_path, mdl_dest + '.tar.gz',
                              applicability_path=apred_path)
        else:
            export_smilespred(pred_path, mdl_dest, compress=False)
            export_smilespred(pred_path, mdl_dest + '.tar.gz')



from setuptools import setup, find_packages
import signaturizer

setup(
    name='signaturizer',
    version=signaturizer.__version__,
    author=signaturizer.__author__,
    author_email=signaturizer.__email__,
    description='Generate Chemical Checker signatures from molecules SMILES.',
    long_description=open('README.md').read().strip(),
    long_description_content_type="text/markdown",
    url='http://gitlabsbnb.irbbarcelona.org/packages/signaturizer',
    py_modules=find_packages(),
    packages=find_packages(),
    install_requires=[
        'tensorflow<2',
        'tensorflow_hub',
        'numpy',
        'h5py'],
    zip_safe=False,
    license='MIT License',
    keywords='signaturizer bioactivity signatures chemicalchecker chemoinformatics',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ])



"""Signaturize molecules."""
import os
import h5py
import shutil
import numpy as np
from tqdm import tqdm
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow.compat.v1 as tf
    import tensorflow_hub as hub
    from tensorflow.compat.v1.keras.models import Model
    from tensorflow.compat.v1.keras import Input
try:
    from rdkit import Chem
    from rdkit import RDLogger
    from rdkit.Chem import AllChem
except ImportError:
    raise ImportError("requires RDKit " +
                      "https://www.rdkit.org/docs/Install.html")

tf.logging.set_verbosity(tf.logging.ERROR)


class Signaturizer(object):
    """Signaturizer Class.

    Loads TF-hub module, compose a single model, handle verbosity.
    """

    def __init__(self, model_name,
                 base_url="http://chemicalchecker.com/api/db/getSignaturizer/",
                 version='v1.1', local=False, tf_version='1', verbose=False,
                 applicability=True):
        """Initialize a Signaturizer instance.

        Args:
            model(str): The model to load. Possible values:
                - the model name (the bioactivity space (e.g. "B1") )
                - the model path (the directory containing 'saved_model.pb')
                - a list of models names or paths (e.g. ["B1", "B2", "E5"])
                - 'GLOBAL' to get the global (i.e. horizontally stacked)
                    bioactivity signature.
            base_url(str): The ChemicalChecker getModel API URL.
            version(int): Signaturizer version.
            local(bool): Wethere the specified model_name shoudl be
                interpreted as a path to a local model.
            tf_version(int): The Tesorflow version.
            verbose(bool): If True some more information will be printed.
            applicability(bool): Wether to also compute the applicability of
                each prediction.
        """
        self.verbose = verbose
        self.applicability = applicability
        if not isinstance(model_name, list):
            if model_name.upper() == 'GLOBAL':
                self.model_names = [y + x for y in 'ABCDE' for x in '12345']
            else:
                self.model_names = [model_name]
        else:
            if model_name == ['GLOBAL']:
                self.model_names = [y + x for y in 'ABCDE' for x in '12345']
            else:
                if 'GLOBAL' in model_name:
                    raise Exception('"GLOBAL" model can only be used alone.')
                self.model_names = model_name
        # load modules as layer to compose a new model
        main_input = Input(shape=(2048,), dtype=tf.float32, name='main_input')
        sign_output = list()
        app_output = list()
        as_dict = False
        output_key = 'default'
        if len(self.model_names) == 1:
            as_dict = True
            output_key = None
        for name in self.model_names:
            # build module spec
            if local:
                if os.path.isdir(name):
                    url = name
                    if self.verbose:
                        print('LOADING local:', url)
                else:
                    raise Exception('Module path not found!')
            else:
                url = base_url + '%s/%s' % (version, name)
                if self.verbose:
                    print('LOADING remote:', url)

            sign_layer = hub.KerasLayer(url, signature='serving_default',
                                        trainable=False, tags=['serve'],
                                        output_key=output_key,
                                        signature_outputs_as_dict=as_dict)
            sign_output.append(sign_layer(main_input))

            if self.applicability:
                try:
                    app_layer = hub.KerasLayer(
                        url, signature='applicability',
                        trainable=False, tags=['serve'],
                        output_key=output_key,
                        signature_outputs_as_dict=as_dict)
                    app_output.append(app_layer(main_input))
                except Exception as ex:
                    print('WARNING: applicability predictions not available. '
                          + str(ex))
                    self.applicability = False
        # join signature output and prepare model
        if len(sign_output) > 1:
            sign_output = tf.keras.layers.concatenate(sign_output)
        self.model = Model(inputs=main_input, outputs=sign_output)
        # same for applicability
        if self.applicability:
            if len(app_output) > 1:
                app_output = tf.keras.layers.concatenate(app_output)
            self.app_model = Model(inputs=main_input, outputs=app_output)
        # set rdKit verbosity
        if self.verbose:
            RDLogger.EnableLog('rdApp.*')
        else:
            RDLogger.DisableLog('rdApp.*')

    def predict(self, molecules, destination=None, chunk_size=1000,
                keytype='SMILES'):
        """Predict signatures for given SMILES.

        Perform signature prediction for input SMILES. We recommend that the
        list is sorted and non-redundant, but this is optional. Some input
        SMILES might be impossible to interpret, in this case, no prediction
        is possible and the corresponding signature will be set to NaN.

        Args:
            molecules(list): List of strings representing molecules. Can be
                SMILES (by default) or InChI.
            destination(str): File path where to save predictions.
            chunk_size(int): Perform prediction on chunks of this size.
            keytype(str): Wether to interpret molecules as InChI or SMILES.
        Returns:
            results: `SignaturizerResult` class. The ordering of input SMILES
                is preserved.
        """
        # convert input molecules to InChI
        inchies = list()
        if keytype.upper() == 'SMILES':
            for smi in molecules:
                if smi == '':
                    smi = 'INVALID SMILES'
                mol = Chem.MolFromSmiles(smi)
                if mol is None:
                    if self.verbose:
                        print("Cannot get molecule from SMILES: %s." % smi)
                    inchies.append('INVALID SMILES')
                    continue
                inchi = Chem.rdinchi.MolToInchi(mol)[0]
                if self.verbose:
                    print('CONVERTED:', smi, inchi)
                inchies.append(inchi)
        else:
            inchies = molecules
        self.inchies = inchies
        # Prepare result object
        features = len(self.model_names) * 128
        results = SignaturizerResult(len(inchies), destination, features)
        results.dataset[:] = self.model_names
        if results.readonly:
            raise Exception(
                'Destination file already exists, ' +
                'delete or rename to proceed.')

        # predict by chunk
        all_chunks = range(0, len(inchies), chunk_size)
        for i in tqdm(all_chunks, disable=not self.verbose):
            chunk = slice(i, i + chunk_size)
            sign0s = list()
            failed = list()
            for idx, inchi in enumerate(inchies[chunk]):
                try:
                    # read molecule
                    inchi = inchi.encode('ascii', 'ignore')
                    if self.verbose:
                        print('READING', inchi, type(inchi))
                    mol = Chem.inchi.MolFromInchi(inchi)
                    if mol is None:
                        raise Exception(
                            "Cannot get molecule from InChI.")
                    info = {}
                    fp = AllChem.GetMorganFingerprintAsBitVect(
                        mol, 2, nBits=2048, bitInfo=info)
                    bin_s0 = [fp.GetBit(i) for i in range(
                        fp.GetNumBits())]
                    calc_s0 = np.array(bin_s0).astype(np.float32)
                except Exception as err:
                    # in case of failure save idx to fill NaNs
                    if self.verbose:
                        print("SKIPPING %s: %s" % (inchi, str(err)))
                    failed.append(idx)
                    calc_s0 = np.full((2048, ),  np.nan)
                finally:
                    sign0s.append(calc_s0)
            # stack input fingerprints and run signature predictor
            sign0s = np.vstack(sign0s)
            preds = self.model.predict(sign0s, batch_size=1)
            # add NaN where SMILES conversion failed
            if failed:
                preds[np.array(failed)] = np.full(features,  np.nan)
            results.signature[chunk] = preds
            # run applicability predictor
            if self.applicability:
                apreds = self.app_model.predict(sign0s, batch_size=1)
                if failed:
                    apreds[np.array(failed)] = np.nan
                results.applicability[chunk] = apreds
        results.close()
        if self.verbose:
            print('PREDICTION complete!')
        failed = np.isnan(results.signature[:, 0])
        results.failed = failed
        if any(failed) > 0:
            print('Some molecules could not be recognized,'
                  ' the corresponding signatures are NaN')
            if self.verbose:
                for idx in np.argwhere(failed).flatten():
                    print(molecules[idx])
        return results

    @staticmethod
    def _clear_tfhub_cache():
        cache_dir = os.getenv('TFHUB_CACHE_DIR')
        if cache_dir is None:
            cache_dir = '/tmp/tfhub_modules/'
        if not os.path.isdir(cache_dir):
            raise Exception('Cannot find tfhub cache directory, ' +
                            'please set TFHUB_CACHE_DIR variable')
        shutil.rmtree(cache_dir)
        os.mkdir(cache_dir)


class SignaturizerResult():
    """SignaturizerResult class.

    Contain result of a prediction.Results are stored in the following
    numpy vector:

        * ``signatures``: Float matrix where each row is a molecule signature.
        * ``applicability``: Float array with applicability score.
        * ``dataset``: List of bioactivity dataset used.
        * ``failed``: Mask for failed molecules.

    If a destination is specified the result are saved in an HDF5 file with
    the same vector available as HDF5 datasets.
    """

    def __init__(self, size, destination, features=128):
        """Initialize a SignaturizerResult instance.

        Args:
            size (int): The number of molecules being signaturized.
            destination (str): Path to HDF5 file where prediction results will
                be saved.
            features (int, optional): how many feature have to be stored.

        """
        self.dst = destination
        self.readonly = False
        if self.dst is None:
            # simple numpy arrays
            self.h5 = None
            self.signature = np.full((size, features), np.nan, order='F',
                                     dtype=np.float32)
            self.applicability = np.full(
                (size, int(np.ceil(features / 128))), np.nan, dtype=np.float32)
            self.dataset = np.full((int(np.ceil(features / 128)),), np.nan,
                                   dtype=h5py.special_dtype(vlen=str))
            self.failed = np.full((size, ), False, dtype=np.bool)
        else:
            # check if the file exists already
            if os.path.isfile(self.dst):
                print('HDF5 file %s exists, opening in read-only.' % self.dst)
                # this avoid overwriting by mistake
                self.h5 = h5py.File(self.dst, 'r')
                self.readonly = True
            else:
                # create the datasets
                self.h5 = h5py.File(self.dst, 'w')
                self.h5.create_dataset(
                    'signature', (size, features), dtype=np.float32)
                self.h5.create_dataset(
                    'applicability', (size, int(np.ceil(features / 128))),
                    dtype=np.float32)
                self.h5.create_dataset(
                    'dataset', (int(np.ceil(features / 128)),),
                    dtype=h5py.special_dtype(vlen=str))
                self.h5.create_dataset(
                    'failed', (size,),
                    dtype=np.bool)
            # expose the datasets
            self.signature = self.h5['signature']
            self.applicability = self.h5['applicability']
            self.dataset = self.h5['dataset']
            self.failed = self.h5['failed']

    def close(self):
        if self.h5 is None:
            return
        self.h5.close()
        # leave it open for reading
        self.h5 = h5py.File(self.dst, 'r')
        # expose the datasets
        self.signature = self.h5['signature']
        self.applicability = self.h5['applicability']
        self.dataset = self.h5['dataset']
        self.failed = self.h5['failed']
